using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;

namespace FallingBlocks
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        //Textures
        private Texture2D mIntro;
        private Texture2D mMessageBox;
        private Texture2D mBackground;
        private Texture2D mBlock;

        //Game objects
        private Shape mCurrentShape = new Shape();
        private Shape mNextShape = new Shape();
        private Playfield mPlayfield = new Playfield();
        private SpriteFont mText;

        //Keyboard Input
        private float mAutoFallTimer;
        private float mDownTime = 0.3F;
        private float mLeftTime = 0.3F;
        private float mRightTime = 0.3F;
        private float mRotateTime = 0.3F;
        private bool mPausePressed = false;

        //Color to tint the blocks and background when the game is paused
        private Color mPauseColor = Color.MediumSlateBlue;

        //Game data tracking for the current score, total number of lines cleared and the current level
        private int mScore = 0;
        private int mTotalLines = 0;
        private int mLevel = 0;
        private int mNumberOfLinesCleared = 0;

        //Current tint for the blocks based on the current level
        private Color mBlockColor = Color.White;
        private float mAutoFallThreshold = 1.0F;


        //Define the various states the game can be in
        private enum GameState
        {
            Intro,
            Playing,
            Paused,
            GameOver
        }//end enum GameState

        //Stores the current state of the game
        private GameState mCurrentState = GameState.Intro;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            graphics.PreferredBackBufferHeight = 768;
            graphics.PreferredBackBufferWidth = 1024;
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            //Initialize the graphics, the game and the text objects
            InitializeGraphics();
            InitializeGame();
            InitializeText();
        }

        //Initialize the SpriteBatch and textures
        protected void InitializeGraphics()
        {
            mBackground = Content.Load<Texture2D>("Images/Background");
            mBlock = Content.Load<Texture2D>("Images/blocks");
            mMessageBox = Content.Load<Texture2D>("Images/MessageBox");
            mIntro = Content.Load<Texture2D>("Images/IntroScreen");
        }//end protected void InitializeGraphics()

        //Initialize the fonts to be used
        protected void InitializeText()
        {
            mText = Content.Load<SpriteFont>("myFont");
        }//end protected void InitializeText()

        //Initialize the Game
        protected void InitializeGame()
        {
            //Initialize the game data
            mTotalLines = 0;
            mScore = 0;
            mLevel = 0;
            mNumberOfLinesCleared = 0;

            //Clear the playfield and put the first shape into play
            mPlayfield.ClearPlayfield();
            CreateNewShape();

        }//end protected void InitializeGame()


        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // TODO: Add your update logic here
            //The time since Update was called last
            float elapsed = (float)gameTime.ElapsedGameTime.TotalSeconds;

            //Get the current state of the keyboard (what keys are and are not being pressed)
            KeyboardState aKeyboard = Keyboard.GetState();

            //If the Escape key has been pressed, then exit the game
            if (aKeyboard.IsKeyDown(Keys.Escape) == true)
            {
                this.Exit();
            }//end if (aKeyboard.IsKeyDown(Keys.Escape) == true)

            //Update the game objects
            UpdateGameObjects(elapsed, aKeyboard);

            base.Update(gameTime);
        }

        //Update the game objects
        protected void UpdateGameObjects(float elapsed, KeyboardState theKeyboard)
        {
            if (mCurrentState == GameState.Playing)
            {
                //Automatically move the current piece down towards the bottom of the playing area as time passes
                AutoFall(elapsed);

                //Update the playfield, checking for completed lines 
                UpdatePlayfield();

                //Update the current level (if appropriate)
                UpdateLevel();

                //Move the current piece to the left
                if (theKeyboard.IsKeyDown(Keys.Left) == true)
                {
                    mLeftTime += elapsed;
                    if (mLeftTime > 0.1F)
                    {
                        mLeftTime = 0F;

                        mCurrentShape.PositionX -= 1;
                        if (mPlayfield.CheckCollision(mCurrentShape) == true)
                        {
                            mCurrentShape.PositionX += 1;
                        }
                    }
                }
                else
                {
                    mLeftTime = 0.3F;
                }

                //Move the current piece to the right
                if (theKeyboard.IsKeyDown(Keys.Right) == true)
                {
                    mRightTime += elapsed;
                    if (mRightTime > 0.1F)
                    {
                        mRightTime = 0F;

                        mCurrentShape.PositionX += 1;
                        if (mPlayfield.CheckCollision(mCurrentShape) == true)
                        {
                            mCurrentShape.PositionX -= 1;
                        }
                    }
                }
                else
                {
                    mRightTime = 0.3F;
                }

                //Rotate the piece
                if (theKeyboard.IsKeyDown(Keys.Up) == true)
                {
                    mRotateTime += elapsed;
                    if (mRotateTime > 0.1F)
                    {
                        mRotateTime = 0F;

                        mCurrentShape.Rotation += 1;
                        if ((mCurrentShape.Rotation > 3))
                        {
                            mCurrentShape.Rotation = 0;
                        }
                        if (mPlayfield.CheckCollision(mCurrentShape) == true)
                        {
                            mCurrentShape.Rotation -= 1;
                            if ((mCurrentShape.Rotation < 0))
                            {
                                mCurrentShape.Rotation = 3;
                            }
                        }
                        else
                        {
 
                        }
                    }
                }
                else
                {
                    mRotateTime = 0.3F;
                }

                //Move the current piece down
                if (theKeyboard.IsKeyDown(Keys.Down) == true)
                {
                    mDownTime += elapsed;
                    if (mDownTime > 0.1F)
                    {
                        mCurrentShape.PositionY += 1;
                        mDownTime = 0F;

                        if (mPlayfield.CheckCollision(mCurrentShape) == true)
                        {
                            mCurrentShape.PositionY -= 1;
                        }
                    }
                }
                else
                {
                    mDownTime = 0.3F;
                }
            }

            if (theKeyboard.IsKeyDown(Keys.P) == true)
            {
                if (mCurrentState == GameState.Paused | mCurrentState == GameState.Playing)
                {
                    if (mPausePressed == false)
                    {
                        mPausePressed = true;
                        if (mCurrentState == GameState.Paused)
                        {
                            mCurrentState = GameState.Playing;
                        }
                        else
                        {
                            mCurrentState = GameState.Paused;
                        }
                    }
                }
            }
            else
            {
                mPausePressed = false;
            }

            if (theKeyboard.IsKeyDown(Keys.Space) == true)
            {
                if (mCurrentState == GameState.Intro)
                {
                    mCurrentState = GameState.Playing;
                }
                else if (mCurrentState == GameState.GameOver)
                {
                    InitializeGame();
                    mCurrentState = GameState.Playing;
                }
            }

        }//end protected void UpdateGameObjects(float elapsed, KeyboardState theKeyboard)

        //Update the game to the next level if appropriate
        protected void UpdateLevel()
        {
            //Update the level if the number of lines cleared is great enough to move to the next level
            if (mTotalLines >= (mLevel + 1) * 10)
            {
                mLevel += 1;
                mBlockColor = GetBlockColor();
                mAutoFallThreshold -= .1F;
            }//end if (mTotalLines >= (mLevel + 1) * 10)
        }//end protected void UpdateLevel()

        //Update the playfield area, clearing lines, updating score and playing sounds 
        protected void UpdatePlayfield()
        {
            //Check for full rows
            int aNumberOfLinesCleared = mPlayfield.CheckRows();
            if (aNumberOfLinesCleared > 0)
            {
                mNumberOfLinesCleared += 1;
                mTotalLines += 1;
            }
            else
            {
                if (mNumberOfLinesCleared >= 3)
                {
                    mScore += 1000;
                }
                else if (mNumberOfLinesCleared > 0)
                {
                    mScore += 100 * mNumberOfLinesCleared;
                }//end if (mNumberOfLinesCleared >= 3)

                mNumberOfLinesCleared = 0;
            }//end if (aNumberOfLinesCleared > 0)
        }//end protected void UpdatePlayfield()

        //Get the color tint for the blocks for the current level
        private Color GetBlockColor()
        {
            Color mColor = new Color((byte)(RandomNumber(0, 255)), (byte)(RandomNumber(0, 255)), (byte)(RandomNumber(0, 255)));

            switch (mLevel)
            {
                case 0:
                    {
                        mColor = Color.White;
                        break;
                    }

                case 1:
                    {
                        mColor = Color.Pink;
                        break;
                    }

                case 2:
                    {
                        mColor = Color.Blue;
                        break;
                    }

                case 3:
                    {
                        mColor = Color.Wheat;
                        break;
                    }

                case 4:
                    {
                        mColor = Color.Red;
                        break;
                    }

                case 5:
                    {
                        mColor = Color.SeaGreen;
                        break;
                    }

                case 6:
                    {
                        mColor = Color.RosyBrown;
                        break;
                    }

                case 7:
                    {
                        mColor = Color.Aqua;
                        break;
                    }

                case 8:
                    {
                        mColor = Color.BlanchedAlmond;
                        break;
                    }

                case 9:
                    {
                        mColor = Color.DarkKhaki;
                        break;
                    }

                case 10:
                    {
                        mColor = Color.Fuchsia;
                        break;
                    }
            }

            return mColor;
        }//end private Color GetBlockColor()

        //Description: Automatically move the current piece down towards the bottom of the playing area as time passes
        private void AutoFall(float elapsedSeconds)
        {
            mAutoFallTimer += elapsedSeconds;
            if (mAutoFallTimer >= mAutoFallThreshold)
            {
                mCurrentShape.PositionY += 1;
                mAutoFallTimer = 0;

                if (mPlayfield.CheckCollision(mCurrentShape) == true)
                {
                    if (mCurrentShape.PositionY <= 0)
                    {
                        mCurrentState = GameState.GameOver;
                        return;
                    }//end if (mCurrentShape.PositionY <= 0)

                    //Add the shape to the playfield
                    mPlayfield.AddShape(mCurrentShape);

                    //Create a new shape
                    CreateNewShape();
                }//end if (mPlayfield.CheckCollision(mCurrentShape) == true)
            }//end if (mAutoFallTimer >= 1)
        }//end private void AutoFall(float elapsedSeconds)

        //Description: Create a new random shape 
        private void CreateNewShape()
        {
            if (mNextShape.Type == Shape.ShapeType.NotSet)
            {
                mNextShape.Type = ((Shape.ShapeType)(RandomNumber(0, 6)));
            }//end if (mNextShape.Type == Shape.ShapeType.NotSet)

            mCurrentShape.Type = mNextShape.Type;
            mNextShape.Type = ((Shape.ShapeType)(RandomNumber(0, 6)));

            mCurrentShape.PositionX = 5;
            mCurrentShape.PositionY = 0;
            mCurrentShape.Rotation = 0;
        }//end private void CreateNewShape()

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            graphics.GraphicsDevice.Clear(Color.CornflowerBlue);

            spriteBatch.Begin();

            if (mCurrentState == GameState.Intro)
            {
                spriteBatch.Draw(mIntro, new Rectangle(0, 0, 1024, 768), Color.White);
            }
            else
            {
                spriteBatch.Draw(mBackground, new Rectangle(0, 0, 1024, 768), BackgroundColor);
                mCurrentShape.Draw(350 + (Convert.ToInt32(mCurrentShape.BlockSize) * mCurrentShape.PositionX), 10 + (Convert.ToInt32(mCurrentShape.BlockSize) * mCurrentShape.PositionY), BlockColor, spriteBatch, mBlock);
                mNextShape.Draw(810, 60, BlockColor, spriteBatch, mBlock);
                mPlayfield.Draw(spriteBatch, mBlock, BlockColor);

                if (mCurrentState == GameState.Paused | mCurrentState == GameState.GameOver)
                {
                    spriteBatch.Draw(mMessageBox, new Rectangle(300, 200, 500, 300), Color.White);
                }
            }

            DrawText(spriteBatch);

            spriteBatch.End();

            base.Draw(gameTime);
        }

        //Determine the current color to "tint" the background
        protected Color BackgroundColor
        {
            get
            {
                Color mColor = Color.White;
                if (mCurrentState == GameState.Paused)
                {
                    mColor = mPauseColor;
                }

                return mColor;
            }
        }//end BackgroundColor

        //Determine the current color to "tint" the blocks
        protected Color BlockColor
        {
            get
            {
                Color mColor = Color.White;

                if (mCurrentState == GameState.Paused)
                {
                    mColor = mPauseColor;
                }
                else
                {
                    mColor = mBlockColor;
                }

                return mColor;
            }
        }//end BlockColor

        //Draw the text to be displayed
        protected void DrawText(SpriteBatch theBatch)
        {
            if (mCurrentState == GameState.Intro)
            {
                spriteBatch.DrawString(mText, "Press 'Space' to begin!", new Vector2(400, 700), Color.White);
            }
            else
            {
                spriteBatch.DrawString(mText, "Level: " + (mLevel + 1).ToString(), new Vector2(90, 230), Color.White);
                spriteBatch.DrawString(mText, "Lines: " + mTotalLines.ToString(), new Vector2(90, 270), Color.White);
                spriteBatch.DrawString(mText, "Score: " + mScore.ToString(), new Vector2(90, 310), Color.White);

                if (mCurrentState == GameState.GameOver)
                {
                    spriteBatch.DrawString(mText, "Game Over", new Vector2(500, 300), Color.White);
                    spriteBatch.DrawString(mText, "Press 'Space' to play again, 'Esc' to exit.", new Vector2(390, 350), Color.White);
                }//end if (mCurrentState == GameState.GameOver)

                if (mCurrentState == GameState.Paused)
                {
                    spriteBatch.DrawString(mText, "Paused!", new Vector2(510, 300), Color.White);
                    spriteBatch.DrawString(mText, "Press 'P' to un-pause the game.", new Vector2(420, 350), Color.White);
                }//end if (mCurrentState == GameState.Paused)
            }//end if (mCurrentState == GameState.Intro)
        }//end protected void DrawText()

        //Generate a random number between (and including) passed in min and max values
        private Random mRandom = new Random();
        protected int RandomNumber(int min, int max)
        {
            return mRandom.Next(min, max + 1);
        }//end protected int RandomNumber(int min, int max)
    }
}
